/*
  INF3105 - Structures de données et algorithmes
  UQAM / Département d'informatique
  Liste doublement chaînée.
*/

#ifndef __LISTE_H__
#define __LISTE_H__

#include <cassert>

template<class T>
class Liste {
 public:
  class Iterateur;

  Liste();
  ~Liste();

  bool estVide() const;
  void vider();

  bool operator==(const Liste&) const;
  const Liste& operator=(const Liste&);

  T& operator[](const Iterateur&);
  const T& operator[](const Iterateur&) const;

  // Les fonctions insérer retournent un Itérateur sur la position de l'objet inséré.
  // La fonction enlever retourne un Iterateur sur la position suivante de l'objet enlevé.
  Iterateur inserer(const T&, const Iterateur&);
  Iterateur enlever(const Iterateur&);

  Iterateur inserer_debut(const T&);
  Iterateur inserer_fin(const T&);
  void enlever_debut();

  Iterateur debut() const;  // Iterateur sur le début.
  Iterateur fin() const;    // Iterateur sur la fin. La fin est la position suivante du dernier élément.

  Iterateur trouver( const T&) const;  // Retourne un itérateur sur le premier élément cherché.
                                       // S'il n'existe pas, retourne fin().

 private:
  class Cellule {
   public:
    Cellule(const T& c, Cellule* p = nullptr, Cellule* s = nullptr)
        : contenu(c), precedente(p), suivante(s) {}
    T contenu;
    Cellule* precedente;
    Cellule* suivante;
  };

 public:
  class Iterateur {
   public:
    Iterateur(const Liste&);
    Iterateur(const Iterateur&);
    Iterateur(const Liste&, Cellule*);

    Iterateur& operator=(int);

    operator bool() const;  // retourne vrai si l'itérateur n'est pas rendu à la fin. Permet de faire while(iter)
    bool operator!() const;  // retourne faux si l'itérateur n'est pas rendu à la fin
    bool operator==( const Iterateur&) const;  // compare deux itérateur (deux "positions")
    bool operator!=( const Iterateur&) const;  // !compare deux itérateur (deux "positions")

    Iterateur operator++(int);  // post-incrément (i++)
    Iterateur& operator++();    // pré-incrément (++i)

    // T& operator*(); // Bonne idée? Pourquoi? Qu'arrive-t-il si la liste est constante?
    const T& operator*() const;  // déréférencement

    Iterateur operator+(int) const;
    Iterateur& operator+=(int);
    Iterateur& operator=(const Iterateur&);

   private:
    Cellule* courante;   // pointeur sur la Cellule courante
    const Liste& liste;  // référence sur la Liste associée à l'itérateur courant (pour des fins de robustesse).

    friend class Liste;
  };

 private:
  Cellule* premiere;
  Cellule* derniere;
};

/*
 Puisque Liste<T> est une classe générique, toutes ses définitions doivent être
 inclues après les déclarations. On peut quand même séparer les déclarations et
 les définitions en deux fichiers (liste.h et liste.hcc), à condition d'inclure
 le deuxième (liste.hcc) à la fin du premier (liste.h). Ainsi, un fichier source
 (.cc, .cpp, c++) qui inclut liste.h inclura indirectement aussi liste.hcc.
*/

template<class T>
Liste<T>::Liste() : premiere(nullptr), derniere(nullptr) {}

template<class T>
Liste<T>::~Liste() {
  vider();
}

template<class T>
bool Liste<T>::estVide() const {
  return premiere == nullptr;
}

template<class T>
void Liste<T>::vider() {
  for (Cellule* c = premiere; c != nullptr; c = c->suivante)
    delete c;

  premiere = nullptr;
  derniere = nullptr;
}

template<class T>
const Liste<T>& Liste<T>::operator=(const Liste<T>& autre) {
  if (this == &autre) return *this;

  vider();
  for (Iterateur ia = autre.debut(); ia; ++ia) inserer_fin(*ia);

  return *this;
}

template<class T>
T& Liste<T>::operator[](const Iterateur& i) {
  assert(&i.liste == this);
  assert(i.courante != nullptr);

  return i.courante->contenu;
}

template<class T>
const T& Liste<T>::operator[](const Iterateur& i) const {
  assert(&i.liste == this);
  assert(i.courante != nullptr);

  return i.courante->contenu;
}

template<class T>
typename Liste<T>::Iterateur Liste<T>::inserer(const T& e, const Iterateur& i) {
  assert(&i.liste == this);
  if (i.courante == nullptr) return inserer_fin(e);
  if (i.courante->precedente == nullptr) return inserer_debut(e);

  Cellule* c = new Cellule(e, i.courante->precedente, i.courante);
  i.courante->precedente->suivante = c;
  i.courante->precedente = c;

  return Iterateur(*this, c);
}

template<class T>
typename Liste<T>::Iterateur Liste<T>::inserer_debut(const T& e) {
  Cellule* c = new Cellule(e, nullptr, premiere);
  if (premiere != nullptr) premiere->precedente = c;
  premiere = c;
  if (derniere == nullptr) derniere = c;

  return Iterateur(*this, c);
}

template<class T>
typename Liste<T>::Iterateur Liste<T>::inserer_fin(const T& e) {
  Cellule* c = new Cellule(e, derniere, nullptr);
  if (derniere != nullptr) derniere->suivante = c;
  derniere = c;
  if (premiere == nullptr) premiere = c;

  return Iterateur(*this, c);
}

template<class T>
typename Liste<T>::Iterateur Liste<T>::enlever(const Iterateur& i) {
  assert(&i.liste == this);
  assert(i.courante != nullptr);

  Cellule* c = i.courante;
  Iterateur position(*this, c->suivante);

  if (c->precedente != nullptr)
    c->precedente->suivante = c->suivante;
  else
    premiere = c->suivante;

  if (c->suivante != nullptr)
    c->suivante->precedente = c->precedente;
  else
    derniere = c->precedente;

  delete c;
  return position;
}

template<class T>
void Liste<T>::enlever_debut() {
  enlever(debut());
}

template<class T>
typename Liste<T>::Iterateur Liste<T>::debut() const {
  return Iterateur(*this, premiere);
}

template<class T>
typename Liste<T>::Iterateur Liste<T>::fin() const {
  return Iterateur(*this, nullptr);
}

template<class T>
typename Liste<T>::Iterateur Liste<T>::trouver(const T& e) const {
  Iterateur i = debut();
  while (i && !(*i == e)) ++i;
  return i;
}

template<class T>
bool Liste<T>::operator==(const Liste& autre) const {
  if (this == &autre) return true;

  Iterateur i1 = debut();
  Iterateur i2 = autre.debut();
  while (i1 && i2) {
    if (*i1 != *i2) return false;
    ++i1;
    ++i2;
  }
  if (i1 && !i2) return false;
  if (!i1 && i2) return false;
  return true;
}

template<class T>
Liste<T>::Iterateur::Iterateur(const Iterateur& i) : liste(i.liste) {
  courante = i.courante;
}

template<class T>
Liste<T>::Iterateur::Iterateur(const Liste<T>& l) : liste(l) {
  courante = liste.premiere;
}

template<class T>
Liste<T>::Iterateur::Iterateur(const Liste<T>& l, Cellule* c) : liste(l) {
  courante = c;
}

template<class T>
Liste<T>::Iterateur::operator bool() const {
  return courante != nullptr;
}

template<class T>
bool Liste<T>::Iterateur::operator!() const {
  return courante == nullptr;
}

template<class T>
bool Liste<T>::Iterateur::operator==(const Iterateur& autre) const {
  assert(&liste == &autre.liste);
  return courante == autre.courante;
}

template<class T>
bool Liste<T>::Iterateur::operator!=(const Iterateur& autre) const {
  assert(&liste == &autre.liste);
  return courante != autre.courante;
}

template<class T>
typename Liste<T>::Iterateur Liste<T>::Iterateur::operator++(int) {
  Iterateur temp(*this);
  if (courante != nullptr) courante = courante->suivante;
  return temp;
}

template<class T>
typename Liste<T>::Iterateur& Liste<T>::Iterateur::operator++() {
  if (courante != nullptr) courante = courante->suivante;
  return *this;
}

template<class T>
const T& Liste<T>::Iterateur::operator*() const {
  assert(courante != nullptr);
  return courante->contenu;
}

template<class T>
typename Liste<T>::Iterateur& Liste<T>::Iterateur::operator+=(int n) {
  for (int i = 0; i < n; ++i)
    operator++();
  return *this;
}

template<class T>
typename Liste<T>::Iterateur Liste<T>::Iterateur::operator+(int n) const {
  Iterateur temp(*this);
  temp += n;
  return temp;
}

template<class T>
typename Liste<T>::Iterateur& Liste<T>::Iterateur::operator=(int n) {
  courante = liste.premiere;
  *this += n;
  return *this;
}

template<class T>
typename Liste<T>::Iterateur& Liste<T>::Iterateur::operator=( const Iterateur& autre) {
  assert(&liste == &autre.liste);

  courante = autre.courante;
  return *this;
}

#endif
